/*
 *  dynamical semantics and general iterated function display tool
 *  based on work by Patrick Grim and Gary Mar
 *  see _The Philosophical Computer_, 1998
 *  Copyright (c) 1999 Matt Chisholm
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <png.h>

typedef struct iters{
  double * x; 
  double * y; 
  double * z; 
  int num_iters;
  double init_x; 
  double init_y;
  double init_z; 
  char * fcn;
  int dims; 
  int width; 
  int height; 
  int splatter;
  char * filename; 
} Iters; 

Iters * iterate_1d_point(Iters * iters){
  int i;
  double * x; 
  double * y; 
  x = new double[iters->num_iters+1];   
  y = new double[iters->num_iters+1]; 
  x[0] = iters->init_x; 
  if (strcmp(iters->fcn, "liar") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = 1-fabs(0-x[i]); 
    }
  } else if (strcmp(iters->fcn, "hs") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = 1-fabs(x[i]/2-x[i]); 
    }
  } else if (strcmp(iters->fcn, "min") == 0){
    for (i = 0; i < iters->num_iters; i++){
      if (x[i] > 0.5){ 
	x[i+1] = 1-fabs(1-2*x[i]); 
      } else if (x[i] < 0.5) {
	x[i+1] = 1-fabs((x[i])-x[i]);
      } else {
	x[i+1] = 1;
      }
    }
  } else if (strcmp(iters->fcn, "el") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = (1-fabs(0-x[i]))*(1-fabs(0-x[i])); 
    }
  } else if (strcmp(iters->fcn, "ml") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = sqrt(1-fabs(0-x[i])); 
    }
  } else if (strcmp(iters->fcn, "cl") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = (1-fabs((1-x[i])-x[i])); 
    }
  } else if (strcmp(iters->fcn, "log") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = (1-(1-fabs((1-x[i])-x[i])))*(1-(1-fabs((1-x[i])-x[i]))); 
    }
  } else if (strcmp(iters->fcn, "nv") == 0){
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = (1-fabs((1-x[i])-x[i]))*(1-fabs((1-x[i])-x[i])); 
    }
  }
  iters->x = x; 
  iters->y = y;
  return iters;
}

//this is the only 2d-iteration function in use
Iters * iterate_2d_point(Iters * iters){
  int i;
  iters->x[0] = iters->init_x;   
  iters->y[0] = iters->init_y; 
  if (strcmp(iters->fcn, "liar") == 0){
    //the liar
    // printf("The Simple Liar\n"); 
    for (i = 0; i < iters->num_iters; i++){
      iters->x[i+1] = iters->y[i]; 
      iters->y[i+1] = iters->x[i]; 
    }
  }else if (strcmp(iters->fcn, "cd") == 0){
    //chaotic dualist
    //printf("The Chaotic Dualist\n"); 
    for (i = 0; i < iters->num_iters; i++){
      iters->x[i+1] = 1-fabs(iters->y[i]-iters->x[i]); 
      iters->y[i+1] = 1-fabs((1-iters->x[i])-iters->y[i]); 
    }
  }else if (strcmp(iters->fcn, "hd") == 0){
    //half dualist
    //printf("The Half Dualist\n"); 
    for (i = 0; i < iters->num_iters; i++){
      iters->x[i+1] = 1-fabs(iters->y[i]/2.0 - iters->x[i]); 
      iters->y[i+1] = 1-fabs((1-iters->x[i+1])-iters->y[i]); 
    }
  }else if (strcmp(iters->fcn, "scd") == 0){
    //successive chaotic dualist - origami
    //printf("The Successive Chaotic Dualist or Origami\n"); 
    for (i = 0; i < iters->num_iters; i++){
      iters->x[i+1] = 1-fabs(iters->y[i]-iters->x[i]); 
      iters->y[i+1] = 1-fabs((1-iters->x[i+1])-iters->y[i]); 
    }
  }else if (strcmp(iters->fcn, "fd") == 0){
    //fuzzy dualist
    //printf("The Fuzzy Dualist\n"); 
    for (i = 0; i < iters->num_iters; i++){
      iters->x[i+1] = (1-(1-fabs(iters->y[i]-iters->x[i])))*(1-(1-fabs(iters->y[i]-iters->x[i]))); 
      iters->y[i+1] = sqrt(1-(1-fabs((1-iters->x[i])-iters->y[i]))); 
    }
  }else if (strcmp(iters->fcn, "sfd") == 0){
    //successive fuzzy dualist
    //printf("The Successive Fuzzy Dualist\n"); 
    for (i = 0; i < iters->num_iters; i++){
      iters->x[i+1] = (1-(1-fabs(iters->y[i]-iters->x[i])))*(1-(1-fabs(iters->y[i]-iters->x[i]))); 
      iters->y[i+1] = sqrt(1-(1-fabs((1-iters->x[i+1])-iters->y[i]))); 
    }
  } else {
    for (i = 0; i < iters->num_iters; i++){
      //printf("Bad user. No biscuit.\n"); 
      iters->x[i+1] = iters->y[i]; 
      iters->y[i+1] = iters->x[i]; 
    }
  }
  return iters;
}

Iters * iterate_3d_point(Iters * iters){
  int i;
  double * x; 
  double * y;
  double * z; 
  x = new double[iters->num_iters+1]; 
  y = new double[iters->num_iters+1];
  z = new double[iters->num_iters+1]; 
  x[0] = iters->init_x;   
  y[0] = iters->init_y; 
  z[0] = iters->init_z; 
  if (strcmp(iters->fcn, "minerva") == 0){
    //minerva
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = 1-fabs(x[i]-(fabs(y[i]-z[i]))/2);
      y[i+1] = 1-fabs(y[i]-(fabs(z[i]-x[i]))/2);
      z[i+1] = 1-fabs(z[i]-(fabs(x[i]-y[i]))/2);
    }
  } else if (strcmp(iters->fcn, "tb") == 0){
    //thunderbird
    for (i = 0; i < iters->num_iters; i++){
      x[i+1] = 1-fabs(x[i]-(fabs(y[i]-z[i]))/4);
      y[i+1] = 1-fabs(y[i]-(fabs(z[i]-x[i]))/4);
      z[i+1] = 1-fabs(z[i]-(fabs(x[i]-y[i]))/4);
    }
  }
  iters->x = x; 
  iters->y = y;
  iters->z = z;
  return iters;
}

int make_png(unsigned char ** row_array, Iters * iters) {
  // variable definitions
  png_structp png_ptr;	  // don't ask, don't tell
  png_infop info_ptr;     // info about png file
  FILE * png_file;        // filehandle
  int image_bits = 8;     // image bits per color
  int image_channels = 3; // channels for image, 3 = RGB

  int width = iters->width; 
  int height = iters->height; 
  char * filename = iters->filename;
  //  filename = "out.png";
  // creates file
  png_file = fopen(filename, "wb");

  //some sort of error check
  if(!png_file) {
      fprintf(stderr, "couldn't open file %s\n", filename);
      exit(-1);
  }
  
  // who knows what this does... assigns values to png_ptr, info_ptr
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  info_ptr = png_create_info_struct(png_ptr);
 
  //make sure we created info_ptr correctly. yee haw.
  if(!info_ptr) {
    fprintf(stderr, "failed to create png struct\n");
    exit(-1);
  }

  // checks setjmp, whatever that is
  if(setjmp(png_ptr->jmpbuf)) {
    fprintf(stderr, "error in setjmp.  exiting\n");
    exit(-1);
  }

  // this does all sorts of shit. y3334h. 31337 h4><0r$ 1|\| 7h3 h0u$3.
  png_init_io(png_ptr, png_file);
  png_set_IHDR(png_ptr, info_ptr, width, height,
  	       image_bits, 
  	       (image_channels == 3) ? PNG_COLOR_TYPE_RGB : 
  	       PNG_COLOR_TYPE_GRAY,
  	       PNG_INTERLACE_NONE,
  	       PNG_COMPRESSION_TYPE_DEFAULT,
  	       PNG_FILTER_TYPE_DEFAULT);
  png_write_info(png_ptr, info_ptr); /* writes this shee-ot to disk */
  png_set_swap(png_ptr);	/* only on little endian ;< */
  //set up file for writing
  png_init_io(png_ptr, png_file);
  //write file
  png_write_image(png_ptr, row_array);
  //alternately, write rows
  //png_write_rows(png_ptr, row_array, height);
  //done writing file
  png_write_end(png_ptr, NULL);
  //close file
  fclose(png_file);
}

void printresults(Iters * iters){
  int i; 
  printf("Iterated %d times.\n", iters->num_iters);
  if (iters->dims == 1){
    for (i = 0; i <= iters->num_iters; i++){
      printf("%d: %f \n", i, iters->x[i]); 
    }
  } else if (iters->dims ==2){
    for (i = 0; i <= iters->num_iters; i++){
      printf("%d: %f, %f \n", i, iters->x[i], iters->y[i]); 
    }

  } else if (iters->dims ==3){
  for (i = 0; i <= iters->num_iters; i++){
    printf("%d: %f, %f, %f \n", i, iters->x[i], iters->y[i], iters->z[i]); 
  }

  } else {}
}

unsigned char ** make_2d_data(Iters * iters){
  unsigned char ** row_array; 
  int i; 
  //int realwidth = (iters->width)*3;
  int x, y; 
  //make row array
  row_array = new (unsigned char *)[iters->height]; 
  for (i = 0; i < iters->height; i++){
    row_array[i] = new (unsigned char)[iters->width*3];
  }
  //set data points in to image
  for (i = 0; i < iters->num_iters; i++){
    x = (int)(iters->x[i]*(iters->width-1))*3;
    y = (int)(iters->y[i]*(iters->height-1));
    //printf("%d(%d, %d)\t", i, x, y); 
    //if (row_array[x][y] == 0x00){
      // printf("already white!\n"); 
    //}
    row_array[y][x] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+1] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+2] = 0xff/**i/(iters->num_iters)*/;
  }
  return row_array;
}

unsigned char ** makemore2d(unsigned char ** row_array, Iters * iters){
  int i; 
  int x; 
  int y; 
  //skipping initial points so i=1
  for (i = 1; i < iters->num_iters; i++){
    x = (int)(iters->x[i]*(iters->width-1))*3;
    y = (int)(iters->y[i]*(iters->height-1));
    row_array[y][x] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+1] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+2] = 0xff/**i/(iters->num_iters)*/;
  }
  return row_array; 
}

unsigned char ** make_1d_data(Iters * iters){
  unsigned char ** row_array; 
  int i; 
  int x, y; 
  //make row array
  row_array = new (unsigned char *)[iters->height]; 
  for (i = 0; i < iters->height; i++){
    row_array[i] = new (unsigned char)[iters->width*3];
  }
  //make web diagram diagonal
  for (i = 0; i < iters->width; i++){
    row_array[i][3*i] = 0xff; 
  }
  //and the function line
  if (strcmp(iters->fcn, "liar") == 0){
    for (i = 0; i < iters->width; i++){
      row_array[iters->width-i-1][3*i] = 0xff; 
    }
  } else if (strcmp(iters->fcn, "hs") == 0){
    for (i = 0; i < iters->width; i++){
      row_array[(int)((iters->width-1)*(1-fabs(((double)i/iters->width)/2.0 -((double)i/iters->width))))][3*i] = 0xff;
    }
  } /*else if (strcmp(iters->fcn, "min") == 0){
    for (i = 0; i < iters->num_iters; i++){
      if (x[i] > 0.5){ 
	x[i+1] = 1-fabs(1-2*x[i]); 
      } else if (x[i] < 0.5) {
	x[i+1] = 1-fabs((x[i])-x[i]);
      } else {
	x[i+1] = 1;
      }
    }
    }*/
  else if (strcmp(iters->fcn, "el") == 0){
    for (i = 0; i < iters->width; i++){
      row_array[(int)((iters->width-1)*(1-fabs(0-(double)i/iters->width))*(1-fabs(0-(double)i/iters->width)))][i*3]=0xff;
    }
  } else if (strcmp(iters->fcn, "ml") == 0){
    for (i = 0; i < iters->width; i++){
      row_array[(int)((iters->width-1)*sqrt(1-fabs(0-(double)i/iters->width)))][i*3] = 0xff; 
    }
  } else if (strcmp(iters->fcn, "cl") == 0){
    for (i = 0; i < iters->width; i++){
      row_array[(int)((iters->width-1)*(1-fabs(1-2*(double)i/iters->width)))][i*3] = 0xff; 
    }
  } else if (strcmp(iters->fcn, "log") == 0){
    for (i = 0; i < iters->width; i++){
      //printf("%d\n",(int)((iters->width-1)*(1-(1-fabs(1-2*(double)i/iters->width)))*(1-(1-fabs(1-2*(double)i/iters->width)))) ); 
      row_array[(int)((iters->width-1)*(1-(1-fabs(1-2*(double)i/iters->width)))*(1-(1-fabs(1-2*(double)i/iters->width))))][i*3] = 0xff; 
    }
  } 
  //set data points in to image
  for (i = 0; i < iters->num_iters; i++){
    //horrid
    x = (int)(iters->x[i]*(iters->width-1))*3;
    y = (int)(iters->x[i]*(iters->height-1));
    // printf("%d, %d\t", x, y); 
    row_array[y][x] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+1] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+2] = 0xff/**i/(iters->num_iters)*/;
    //vertigo
    x = (int)(iters->x[i]*(iters->width-1))*3;
    y = (int)(iters->x[i+1]*(iters->height-1));
    row_array[y][x] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+1] = 0xff/**i/(iters->num_iters)*/;
    row_array[y][x+2] = 0xff/**i/(iters->num_iters)*/;
  }
  return row_array;
}

void help(){
  printf("-d [1, 2, or 3] - dimensions 
-i [NUMBER] - number of iterations
-p [x y z] - initial points - 
   *must* be three of these, even in 1-d and 2-d mode. extras are ignored.
-sp [NUMBER] - pixel spacing of splatter points - only in 2d mode, folx. Sorry.
-size [WIDTH] [HEIGHT] - image size
-f function: 
1D: 
liar\tliar
hs\thalf-sayer
min\tminimalist
el\temphatic liar
ml\tmodest liar
cl\tchaotic liar
log\tlogistic liar
nv\tno variance
2D:
liar\tliar
cd\tchaotic-dualist
hd\thalf-dualist
scd\tsuccessive chaotic dualist
fd\tfuzzy dualist
sfd\tsuccessive fuzzy dualist
3D:
minerva\tminerva
tb\tthunderbird
");
}

void main(int argc, char *argv[]){
  int i;
  int dimensions = 2;
  int splatter = 0; 
  int x, y;  
  unsigned char ** row_array;
  Iters * iters = new Iters;
  iters->num_iters = 10; 
  iters->init_x = 1; 
  iters->init_y = 1;
  iters->init_z = 1; 
  iters->dims = 2;
  iters->width = 500; 
  iters->height = 500; 
  iters->filename = "test.png";
  iters->splatter = 10; 
  iters->fcn = "liar"; 
  if ( argc == 1 ){
    printf("Dynsem need at least some of the following arguments: \n");
    help(); 
    exit(0); 
  }
  //read arguments
  for (i = 0; i < argc; i++){
    if (strcmp(argv[i], "-i") == 0){
      iters->num_iters = atoi(argv[++i]); 
    } else if (strcmp(argv[i], "-d") == 0){
      iters->dims = atoi(argv[++i]); 
    } else if (strcmp(argv[i], "-p") == 0){
      iters->init_x = atof(argv[++i]); 
      iters->init_y = atof(argv[++i]);
      iters->init_z = atof(argv[++i]);
    } else if (strcmp(argv[i], "-f") == 0){
      iters->fcn = argv[++i];
    } else if (strcmp(argv[i], "-sp") == 0){
      iters->dims = 2; 
      splatter = 1; 
      iters->splatter = atoi(argv[++i]); 
    } else if (strcmp(argv[i], "-size") == 0){
      iters->width = atoi(argv[++i]); 
      iters->height = atoi(argv[++i]); 
    } else if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-?") == 0)){
      help(); 
      exit(0); 
    } else {
    }
  }//end read arguments
  printf("%s\n", iters->fcn);
  if (splatter == 0){//pointwise iteration
    if (iters->dims == 1){
      //make iters
      iters = iterate_1d_point(iters); 
      //set iterations in to array
      row_array = make_1d_data(iters);
      //make png
      make_png(row_array, iters);
    }else if (iters->dims == 2){
      //make iters
      iters->x = new double[iters->num_iters+1]; 
      iters->y = new double[iters->num_iters+1]; 
      //iterate point
      iters = iterate_2d_point(iters); 
      //set iterations in to array
      row_array = make_2d_data(iters);
      //make png
      make_png(row_array, iters);
    }else if (iters->dims == 3){
      //make iters
      iters = iterate_3d_point(iters);
      //print results
      printresults(iters); 
    }
  } else {//splatter style iteration
    //make row array
    row_array = new (unsigned char *)[iters->height]; 
    for (i = 0; i < iters->height; i++){
      row_array[i] = new (unsigned char)[iters->width*3];
    }
    //make iters
    iters->x = new double[iters->num_iters+1]; 
    iters->y = new double[iters->num_iters+1]; 
    //splatter method, only works for 2-d
    for (x = 0; x < iters->width; x+=iters->splatter){
      for (y = 0; y < iters->height; y+=iters->splatter){
	//set initial conditions
	iters->init_x = (double)x/iters->width; 
	iters->init_y = (double)y/iters->height; 
	//iterate point
	iters = iterate_2d_point(iters); 
	//set iterations in to array
	row_array = makemore2d(row_array, iters);
      }
    }
    //write png
    make_png(row_array, iters);
  }
}
